#include "renderer.h"

static void _renderer_viewport_window_set(Renderer* self);

void
renderer_buffer_init(Renderer* self, RendererBuffer buffer)
{
	ivec2 size;

	window_size_get(self->window, size);

	fbo_init(&self->fbos[buffer]);

	fbo_bind(&self->fbos[buffer]);

	texture_init(&self->fboTextures[buffer], size, NULL);

	texture_bind(&self->fboTextures[buffer]);
	
	fbo_texture_set(&self->fboTextures[buffer]);

	rbo_init(&self->rbos[buffer]);

	rbo_bind(&self->rbos[buffer]);

	rbo_storage_set(&self->rbos[buffer], size);

	fbo_rbo_set(&self->rbos[buffer]);

	rbo_unbind();
	fbo_unbind();
}

void
renderer_buffer_free(Renderer* self, RendererBuffer buffer)
{
	texture_free(&self->fboTextures[buffer]);
	rbo_free(&self->rbos[buffer]);
	fbo_free(&self->fbos[buffer]);
}

static void
_renderer_viewport_window_set(Renderer* self)
{
	ivec4 viewport;

	window_ivec4_get(self->window, viewport);

	glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
}

void
renderer_init(Renderer* self, Window* window, CameraType type, const ivec2* bufferSizes)
{

	memset(self, '\0', sizeof(Renderer));

	self->window = window;

	self->glContext = SDL_GL_CreateContext(self->window->sdl);

	glew_init();

	vao_init(&self->vao);
	vbo_init(&self->vbo, GL_ARRAY_BUFFER, true);
	vbo_init(&self->ebo, GL_ELEMENT_ARRAY_BUFFER, true);

	for (s32 i = 0; i < RENDERER_BUFFER_COUNT; i++)
	{
		vec2 cameraSize;

		renderer_buffer_init(self, (RendererBuffer)i);

		glm_ivec2_copy((s32*)bufferSizes[i], self->bufferSizes[i]);

		cameraSize[0] = bufferSizes[i][0];
		cameraSize[1] = bufferSizes[i][1];

		switch (type)
		{
			case CAMERA_ORTHOGRAPHIC:
				camera_orthographic_init(&self->camera[i], cameraSize);
				break;
			default:
				break;
		}
	}

	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void
renderer_free(Renderer* self)
{
	for (s32 i = 0; i < RENDERER_BUFFER_COUNT; i++)
		renderer_buffer_free(self, (RendererBuffer)i);

	vao_free(&self->vao);
	vbo_free(&self->vbo);
	vbo_free(&self->ebo);

	memset(self, '\0', sizeof(Renderer));
	SDL_GL_DeleteContext(self->glContext);
}

void
renderer_shader_use(Renderer* self, Shader* shader)
{
	shader_use(shader);
}

void
renderer_clear_color_set(Renderer* self, vec4 color)
{
	glClearColor(color[0], color[1], color[2], color[3]);
}

void
renderer_clear(Renderer* self)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void
renderer_present(Renderer* self)
{
	SDL_GL_SwapWindow(self->window->sdl);
}

void
renderer_update(Renderer* self)
{
	_renderer_viewport_window_set(self);
}

void
renderer_buffer_use(Renderer* self, RendererBuffer buffer)
{
	fbo_bind(&self->fbos[buffer]);
}

void
renderer_buffer_unbind(void)
{
	fbo_unbind();
}

void
renderer_buffer_position_from_window_position
(
	Renderer* self, 
	RendererBuffer buffer,
	vec2 windowPosition, 
	vec2 rendererPosition
)
{
	ivec2 windowSize;

	window_size_get(self->window, windowSize);

	rendererPosition[0] = windowPosition[0] / windowSize[0];
	rendererPosition[1] = windowPosition[1] / windowSize[1];

	rendererPosition[0] *= self->camera[buffer].orthographic.size[0];
	rendererPosition[1] *= self->camera[buffer].orthographic.size[1];

	rendererPosition[0] += (self->camera[buffer].orthographic.size[0] / 2) * (1 - self->camera[buffer].zoom);
	rendererPosition[1] += (self->camera[buffer].orthographic.size[1] / 2) * (1 - self->camera[buffer].zoom);

	rendererPosition[0] += self->camera[buffer].position[0];
	rendererPosition[1] += self->camera[buffer].position[1];

}
