#include "level.h"

static void _level_entity_init(Level* self);
static void _level_finish(Level* self);
static void _level_finish_text_add(Level* self);
static void _level_finish_amulet_spawn(Level* self);

static void
_level_finish_text_add(Level* self)
{
	char string[LEVEL_FINISH_STRING_MAX];
	char* medalString;
	char* completeString;
	ComponentFood* targetFood;
	bool isFail;

	isFail = false;

	switch (self->medal)
	{

		case LEVEL_MEDAL_BRONZE:
			medalString = STRING_LEVEL_MEDAL_BRONZE;
			break;
		case LEVEL_MEDAL_SILVER:
			medalString = STRING_LEVEL_MEDAL_SILVER;
			break;
		case LEVEL_MEDAL_GOLD:
			medalString = STRING_LEVEL_MEDAL_GOLD;
			break;
		case LEVEL_MEDAL_NONE:
		default:
			medalString = STRING_LEVEL_MEDAL_NONE;
			isFail = true;
			break;
	}

	if (isFail)
	{
		snprintf
		(
			string, 
			LEVEL_FINISH_STRING_MAX, 
			LEVEL_FINISH_TEXT_FAILED_FORMAT,
			STRING_LEVEL_FAILED,
			medalString,
			self->score
		);
	}
	else
	{
		u32 milliseconds;
		u32 minutes;
		u32 seconds;

		time_formatted_get(self->timeToMedal, &minutes, &seconds, &milliseconds);
		
		snprintf
		(
			string, 
			LEVEL_FINISH_STRING_MAX, 
			LEVEL_FINISH_TEXT_FORMAT,
			STRING_LEVEL_COMPLETE,
			medalString,
			minutes,
			seconds,
			milliseconds,
			self->score
		);
	}
	
	level_text_queue_clear(self);
	level_text_queue_add(self, string);
}

static void
_level_finish(Level* self)
{
	ComponentTimer* finishTimer;
	ComponentFood* playerFood;
	ComponentFood* targetFood;
	ComponentStage* targetStage;

	finishTimer = ecs_component_get(self->ecs, COMPONENT_TIMER, self->finishTimer);
	playerFood = ecs_component_get(self->ecs, COMPONENT_FOOD, self->player);
	targetFood = ecs_component_get(self->ecs, COMPONENT_FOOD, self->target);
	targetStage = ecs_component_get(self->ecs, COMPONENT_STAGE, self->target);

	playerFood->value = 0;

	finishTimer->component.isDisabled = false;

	if (self->settings.isEvent)
	{
		ComponentTimer* eventTimer;
		
		eventTimer = ecs_component_get(self->ecs, COMPONENT_TIMER, self->eventTimer);
		
		eventTimer->component.isDisabled = true;
	}

	self->isTimerDone = true;
	self->score = targetFood->value;

	if (self->medal >= LEVEL_MEDAL_BRONZE)
		self->isComplete = true;

	if (self->isComplete)
	{
		if ((s32)targetStage->value < TARGET_ATLAS_SIZE[0] - 1)
		{
			targetStage->value++;
			sound_play(&self->ecs->resources->sounds[SOUND_GURGLE + random_s32_in_range(0, SOUND_GURGLE_COUNT - 1)]);
		}
	}

	sound_play(&self->ecs->resources->sounds[SOUND_FINISH]);
	music_play(&self->ecs->resources->music[MUSIC_WIND], true);

	entity_text_disappearing_add
	(
		self->ecs,
		&self->ecs->resources->fonts[FONT_BIG],
		RENDERER_BUFFER_UI,
		STRING_LEVEL_FINISH,
		LEVEL_FINISH_TEXT_POSITION,
		-1,
		LEVEL_FINISH_TEXT_VALUE
	);

	level_event_return(self);

	level_collectible_spawn_clear(self);

	_level_finish_text_add(self);
}

static void
_level_finish_amulet_spawn(Level* self)
{
	ComponentPhysics* physics;
	vec3 smokePosition;

	self->amulet = entity_collectible_amulet_add
	(
		self->ecs,
		LEVEL_AMULET_POSITION
	);

	physics = ecs_component_get(self->ecs, COMPONENT_PHYSICS, self->amulet);

	glm_vec3_copy(physics->position, smokePosition);

	smokePosition[2] += 0.001f;

	entity_smoke_big_add
	(
	 	self->ecs,
		smokePosition
	);

}

static void 
_level_entity_init(Level* self)
{
	ComponentTimer* finishTimer;

	self->bg = entity_sprite_add
	(
		self->ecs,
		&self->ecs->resources->textures[TEXTURE_GAME_BG],
		RENDERER_BUFFER_BACKGROUND,
		ORIGIN_TOP_LEFT,
		(f32*)LEVEL_BG_SIZE,
		(f32*)LEVEL_BG_POSITION
	);

	self->plane = entity_sprite_add
	(
		self->ecs,
		&self->ecs->resources->textures[TEXTURE_GAME_PLANE],
		RENDERER_BUFFER_PLANE,
		ORIGIN_TOP_LEFT,
		(f32*)LEVEL_PLANE_SIZE,
		(f32*)LEVEL_PLANE_POSITION
	);

	self->timer = entity_timer_add
	(
		self->ecs,
		self->settings.time
	);

	self->finishTimer = entity_timer_add
	(
		self->ecs,
		LEVEL_FINISH_TIMER_VALUE
	);

	finishTimer = ecs_component_get(self->ecs, COMPONENT_TIMER, self->finishTimer);
	finishTimer->component.isDisabled = true;

	self->player = entity_player_add
	(
		self->ecs,
		LEVEL_PLAYER_POSITION
	);

	self->target = entity_target_add
	(
		self->ecs,
		LEVEL_TARGET_POSITION,
		self->settings.medalThresholds[0],
		self->levelValue,
		self->player
	);
}

void
level_init(Level* self, ECS* ecs, LevelSettings settings, u32 levelValue)
{
	memset(self, '\0', sizeof(Level));

	self->ecs = ecs;
	
	level_pause_init(self);
	level_unpause(self);

	random_seed_set(0);

	self->settings = settings;
	self->medal = LEVEL_MEDAL_NONE;
	self->nextMedal = LEVEL_MEDAL_BRONZE;
	self->timerValue = 0;
	self->levelValue = levelValue;
	self->isCountdownDone = false;

	self->collectibleSeed = self->settings.seedStart;
	self->eventSeed = self->settings.seedStart;

	_level_entity_init(self);

	if (self->settings.isEvent)
		level_event_init(self);

	level_ui_init(self);

	level_text_queue_init(self);
	level_countdown_init(self);

	level_collectible_spawn_init(self);

	if (self->settings.isSunset)
		glm_vec4_copy(LEVEL_SUNSET_COLOR, self->ecs->postprocessing[RENDERER_BUFFER_WORLD].color);
	
	glm_vec4_copy(settings.colorCover, self->ecs->postprocessing[RENDERER_BUFFER_PLANE].colorCover);
	glm_vec4_copy(settings.colorCover, self->ecs->postprocessing[RENDERER_BUFFER_WORLD].colorCover);
	glm_vec4_copy(settings.colorCover, self->ecs->postprocessing[RENDERER_BUFFER_OVERLAY].colorCover);
	glm_vec4_copy(settings.colorCover, self->ecs->postprocessing[RENDERER_BUFFER_BACKGROUND].colorCover);

	music_play(&self->ecs->resources->music[MUSIC_WIND], true);
}

void
level_tick(Level* self)
{
	ComponentTimer* timer;
	ComponentTimer* finishTimer;

	level_collectible_spawn_tick(self);
	level_countdown_tick(self);
	level_text_queue_tick(self);
	level_pause_tick(self);
	
	if (self->settings.isEvent && !self->isTimerDone && self->isCountdownDone)
		level_event_tick(self);

	timer = ecs_component_get(self->ecs, COMPONENT_TIMER, self->timer);

	self->timerValue = timer->value;

	level_ui_tick(self);

	if 
	(
		(!timer->isFinished && timer->value <= LEVEL_TIMER_LOW_TICK_THRESHOLD) &&
		self->isCountdownDone
	)
	{
		ComponentText* text;

		text = ecs_component_get(self->ecs, COMPONENT_TEXT, self->timerUI);

		if (timer->value % SECOND_TICK == 0)
			sound_play(&self->ecs->resources->sounds[SOUND_LOW_TIMER_TICK]);

		glm_vec4_copy(LEVEL_EVENT_TIMER_LOW_TICK_THRESHOLD_COLOR, text->color);
	}

	if (!timer->isFinished)
		return;

	if (timer->isFinished && !self->isTimerDone)
		_level_finish(self);

	if (self->settings.isAmulet)
	{
		if (!self->isAmuletSpawn)
		{
			_level_finish_amulet_spawn(self);	
			self->isAmuletSpawn = true;
		}
		else
		{
			if (self->ecs->lists[COMPONENT_COLLECTIBLE].components.count == 0)
				self->isFinished = true;
		}
	}
	else
	{

		finishTimer = ecs_component_get(self->ecs, COMPONENT_TIMER, self->finishTimer);

		if (finishTimer->isFinished)
			self->isFinished = true;
	}
}

void
level_free(Level* self)
{
	level_text_queue_free(self);
}

